(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory KBO
imports Term_Order Term_Multiset Lexicographic_Extension
begin

locale weight_fun_prc =
  fixes w :: "'f \<times> nat \<Rightarrow> nat"
  and w0 :: nat
  and prc :: "'f \<times> nat \<Rightarrow> nat"
  and least :: "'f \<Rightarrow> bool"
begin

fun weight :: "('f,'v)term \<Rightarrow> nat"
  where "weight (Var x) = w0"
      | "weight (Fun f ts) = w (f,length ts) + listsum (map weight ts)"

fun kbo :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool \<times> bool"
  where "kbo s t = (if (vars_term_ms t \<le> vars_term_ms s \<and> weight t \<le> weight s)
       then if (weight t < weight s) then (True,True)
       else (case s of
      Var y \<Rightarrow> (False, (case t of Var x \<Rightarrow> x = y | Fun g ts \<Rightarrow> ts = [] \<and> least g)) 
    | Fun f ss \<Rightarrow> case t of Var x \<Rightarrow> (True,True)
                 | Fun g ts \<Rightarrow> if prc (f,length ss) > prc (g,length ts) then (True,True) else if prc (f,length ss) \<ge> prc (g,length ts) then 
              lex_ext_unbounded kbo ss ts else (False,False))
       else (False,False))" 

abbreviation S where "S \<equiv> \<lambda> s t. fst (kbo s t)"
abbreviation NS where "NS \<equiv> \<lambda> s t. snd (kbo s t)"

lemma NS_refl: "NS s s"
proof (induct s)
  case (Var x)
  show ?case by simp
next
  case (Fun f ss)
  have "snd (lex_ext kbo (length ss) ss ss)"
    by (rule all_nstri_imp_lex_nstri, insert Fun[unfolded set_conv_nth], auto)
  hence "snd (lex_ext_unbounded kbo ss ss)" unfolding lex_ext_def Let_def by simp
  thus ?case by auto 
qed

end

locale admissible_weight_fun_prc = weight_fun_prc w w0 prc least 
  for w w0 prc and least :: "'f \<Rightarrow> bool" +
  assumes w0: "\<And> f. w (f,0) \<ge> w0" "w0 > 0"
  and adm: "\<And> f g n. w (f,1) = 0 \<Longrightarrow> prc (g,n) \<le> prc (f,1)"
  and least: "\<And> f. least f = (w (f,0) = w0 \<and> (\<forall> g. w (g,0) = w0 \<longrightarrow> prc (f,0) \<le> prc (g,0)))"
begin

lemma weight_w0: "weight t \<ge> w0"
proof (induct t)
  case (Fun f ts)
  show ?case
  proof (cases ts)
    case Nil
    with w0 show ?thesis by auto
  next
    case (Cons s ss)
    hence "weight s \<ge> w0" using Fun by auto
    thus ?thesis unfolding Cons by auto
  qed
qed simp

lemma weight_subst: "weight (t \<cdot> \<sigma>) = weight t + mset_sum (Multiset2.map (\<lambda> x. weight (\<sigma> x) - w0) (vars_term_ms t))"
proof (induct t)
  case (Var x)
  show ?case using weight_w0[of "\<sigma> x"] by auto
next
  case (Fun f ts)
  have id: "map (\<lambda> t. weight (t \<cdot> \<sigma>)) ts = map (\<lambda> t. weight t + mset_sum (Multiset2.map (\<lambda> x. weight (\<sigma> x) - w0) (vars_term_ms t))) ts"
    by (rule map_cong[OF refl Fun])
  show ?case by (simp add: o_def id, induct ts, auto)
qed

lemma weight_stable_le: assumes ws: "weight s \<le> weight t" and 
  vs: "vars_term_ms s \<le> vars_term_ms t"
  shows "weight (s \<cdot> \<sigma>) \<le> weight (t \<cdot> \<sigma>)"
proof -
  from vs[unfolded mset_le_exists_conv] obtain u where vt: "vars_term_ms t = vars_term_ms s + u" ..
  show ?thesis unfolding weight_subst vt using ws by auto
qed

lemma weight_stable_lt: assumes ws: "weight s < weight t" and 
  vs: "vars_term_ms s \<le> vars_term_ms t"
  shows "weight (s \<cdot> \<sigma>) < weight (t \<cdot> \<sigma>)"
proof -
  from vs[unfolded mset_le_exists_conv] obtain u where vt: "vars_term_ms t = vars_term_ms s + u" ..
  show ?thesis unfolding weight_subst vt using ws by auto
qed

lemma not_S_Var: "\<not> S (Var x) t"
  using weight_w0[of t] by (cases t, auto)

lemma S_imp_NS: "S s t \<Longrightarrow> NS s t"
proof (induct s t rule: kbo.induct)
  case (1 s t)
  from 1(2) have S: "S s t" .
  from S have w: "vars_term_ms t \<le> vars_term_ms s \<and> weight t \<le> weight s" 
    by (auto split: if_splits)
  note S = S w
  note IH = 1(1)[OF w]
  show ?case
  proof (cases "weight t < weight s")
    case True
    with S show ?thesis by simp
  next
    case False
    note IH = IH[OF False]
    note S = S False
    from not_S_Var[of _ t] S
    obtain f ss where s: "s = Fun f ss" by (cases s, auto)
    note IH = IH[OF s]
    show ?thesis
    proof (cases t)
      case (Var x)
      from S show ?thesis by (auto, insert Var s, auto)
    next
      case (Fun g ts)
      note IH = IH[OF Fun]      
      let ?f = "(f,length ss)"
      let ?g = "(g,length ts)"
      let ?lex = "lex_ext_unbounded kbo ss ts"
      from S[simplified, unfolded s Fun] have disj: "prc ?f > prc ?g \<or> prc ?f \<ge> prc ?g \<and> fst ?lex" by (auto split: if_splits)
      show ?thesis
      proof (cases "prc ?f > prc ?g")
        case True
        thus ?thesis using S s Fun by auto
      next
        case False
        with disj have fg: "prc ?f \<ge> prc ?g" and lex: "fst ?lex" by auto
        note IH = IH[OF False fg]
        from lex have "fst (lex_ext kbo (length ss + length ts) ss ts)" 
          unfolding lex_ext_def Let_def by auto
        from lex_ext_stri_imp_nstri[OF this] have lex: "snd ?lex" 
          unfolding lex_ext_def Let_def by auto
        with False fg S s Fun show ?thesis by auto
      qed
    qed
  qed
qed

lemma NS_Var_imp_eq_least: "NS (Var x) t \<Longrightarrow> t = Var x \<or> (\<exists> f. t = Fun f [] \<and> least f)"
  by (cases t, insert weight_w0[of t], auto split: if_splits)
    
lemma S_mono_one: assumes S: "S s t" shows "S (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft))"
proof -
  let ?ba = "\<lambda> s. bef @ s # aft"
  let ?ss = "?ba s"
  let ?ts = "?ba t"
  let ?s = "Fun f ?ss"
  let ?t = "Fun f ?ts"
  from S have w: "weight t \<le> weight s" and v: "vars_term_ms t \<le> vars_term_ms s" 
    by (auto split: if_splits)
  from v have v': "vars_term_ms ?t \<le> vars_term_ms ?s" by auto
  from w have w': "weight ?t \<le> weight ?s" by auto
  have lex: "fst (lex_ext_unbounded kbo ?ss ?ts)"
    unfolding lex_ext_unbounded_iff fst_conv
    by (rule disjI1, rule exI[of _ "length bef"], insert S NS_refl, auto simp del: kbo.simps simp: nth_append)    
  show ?thesis using v' w' lex by (simp split: if_splits)
qed

lemma NS_mono_one: assumes NS: "NS s t" shows "NS (Fun f (bef @ s # aft)) (Fun f (bef @ t # aft))"
proof -
  let ?ba = "\<lambda> s. bef @ s # aft"
  let ?ss = "?ba s"
  let ?ts = "?ba t"
  let ?s = "Fun f ?ss"
  let ?t = "Fun f ?ts"
  from NS have w: "weight t \<le> weight s" and v: "vars_term_ms t \<le> vars_term_ms s" 
    by (auto split: if_splits)
  from v have v': "vars_term_ms ?t \<le> vars_term_ms ?s" by auto
  from w have w': "weight ?t \<le> weight ?s" by auto
  have lex: "snd (lex_ext_unbounded kbo ?ss ?ts)"
    unfolding lex_ext_unbounded_iff snd_conv
  proof (intro disjI2 conjI allI impI)
    fix i
    assume "i < length (bef @ t # aft)"
    thus "NS (?ss ! i) (?ts ! i)" using NS NS_refl
      by (cases "i = length bef", auto simp del: kbo.simps simp: nth_append)
  qed simp
  show ?thesis using v' w' lex by (simp split: if_splits)
qed

lemma kbo_supt_one: "NS s t \<Longrightarrow> S (Fun f (bef @ s # aft)) t"
proof (induct t arbitrary: f s bef aft)
  case (Var x)
  note NS = this
  let ?ss = "bef @ s # aft"
  have ws: "weight s \<le> listsum (map weight ?ss)" by auto
  have "Fun f ?ss \<unrhd> s" by auto
  from supteq_imp_vars_term_ms_subset[OF this] 
  have vs: "vars_term_ms s \<le> vars_term_ms (Fun f ?ss)" .
  let ?t = "Var x"
  from NS have wt: "weight ?t \<le> weight s" and 
    vt: "vars_term_ms ?t \<le> vars_term_ms s" by (auto split: if_splits)
  from ws wt have w: "weight ?t \<le> listsum (map weight ?ss)" by simp
  from vt vs have v: "vars_term_ms ?t \<le> vars_term_ms (Fun f ?ss)" by auto
  from w v show ?case by auto
next
  case (Fun g ts f s bef aft)
  let ?t = "Fun g ts"
  let ?ss = "bef @ s # aft"
  note NS = Fun(2)
  note IH = Fun(1)
  have ws: "weight s \<le> listsum (map weight ?ss)" by auto
  have "Fun f ?ss \<unrhd> s" by auto
  from supteq_imp_vars_term_ms_subset[OF this] 
  have vs: "vars_term_ms s \<le> vars_term_ms (Fun f ?ss)" .
  from NS have wt: "weight ?t \<le> weight s" and 
    vt: "vars_term_ms ?t \<le> vars_term_ms s" by (auto split: if_splits)
  from ws wt have w: "weight ?t \<le> listsum (map weight ?ss)" by simp
  from vt vs have v: "vars_term_ms ?t \<le> vars_term_ms (Fun f ?ss)" by auto
  show ?case
  proof (cases "weight ?t = weight (Fun f ?ss)")
    case False
    with w v show ?thesis by auto
  next
    case True
    from wt[unfolded True] 
    have wf: "w (f,length ?ss) = 0"
      and lsum: "listsum (map weight (bef @ aft)) = 0" by auto
    from lsum have "bef @ aft = []" using weight_w0[of "hd (bef @ aft)"] w0(2) 
      by (cases "bef @ aft", auto)
    hence ba: "bef @ s # aft = [s]" by simp
    with wf have wf: "w (f,1) = 0" by auto
    from True[unfolded ba] wf have wst: "weight s = weight ?t" by simp
    let ?g = "prc (g,length ts)"
    let ?f = "prc (f,1)"
    show ?thesis
    proof (cases "?g < ?f")
      case True
      with w v show ?thesis unfolding ba by simp
    next
      case False
      note admf = adm[OF wf]
      from False admf[of g "length ts"] have pg: "?g = ?f" by auto
      note g = admf[unfolded pg[symmetric]]
      show ?thesis
      proof (cases ts)
        case Nil
        have "fst (lex_ext_unbounded kbo [s] ts)" 
          unfolding Nil lex_ext_unbounded_iff by auto
        with pg w v show ?thesis unfolding ba by simp
      next
        case (Cons t tts)
        {
          fix x
          assume s: "s = Var x"          
          from NS_Var_imp_eq_least[OF NS[unfolded s Cons]] have False by auto
        }
        then obtain h ss where s: "s = Fun h ss" by (cases s, auto)
        from NS wst g[of h "length ss"] have lex: "snd (lex_ext_unbounded kbo ss ts)"
          unfolding s by (auto split: if_splits)
        from lex obtain s0 sss where ss: "ss = s0 # sss" unfolding Cons lex_ext_unbounded_iff snd_conv by (cases ss, auto)
        from lex[unfolded ss Cons] have "S s0 t \<or> NS s0 t"
          by (cases "kbo s0 t", simp add: lex_ext_unbounded.simps del: kbo.simps split: if_splits)
        with S_imp_NS[of s0 t] have "NS s0 t" by blast
        from IH[OF _ this, of h Nil sss] have S: "S s t" unfolding Cons s ss by simp
        have lex: "fst (lex_ext_unbounded kbo [s] ts)" unfolding Cons
          unfolding lex_ext_unbounded_iff fst_conv
          by (rule disjI1[OF exI[of _ 0]], insert S, auto simp del: kbo.simps)
        from lex wst S pg v show ?thesis unfolding ba unfolding s ss Cons
          by simp
      qed
    qed
  qed
qed

lemma kbo_supt: assumes supt: "s \<rhd> t" shows "S s t"
proof -
  from supt obtain C where s: "s = C\<langle>t\<rangle>" and C: "C \<noteq> \<box>" by auto
  show ?thesis unfolding s using C
  proof (induct C arbitrary: t)
    case Hole thus ?case by simp
  next
    case (More f bef C aft t)
    show ?case
    proof (cases "C = \<box>")
      case True
      from kbo_supt_one[OF NS_refl, of f bef t aft] show ?thesis unfolding True by simp
    next
      case False      
      from kbo_supt_one[OF S_imp_NS[OF More(1)[OF False]], of f bef t aft]
      show ?thesis by simp
    qed
  qed
qed   

lemma kbo_supteq: assumes supt: "s \<unrhd> t" shows "NS s t"
  using S_imp_NS[OF kbo_supt[of s t]] NS_refl[of s] using supt[unfolded subterm.le_less]
  by blast

lemma NS_all_least: assumes l: "least f"
  shows "NS t (Fun f [])"
proof (induct t)
  case (Var x)
  show ?case using l[unfolded least] l 
    by (auto simp: empty_le)
next
  case (Fun g ts)
  show ?case
  proof (cases ts)
    case (Cons s ss)
    with Fun[of s] have "NS s (Fun f [])" by auto
    from S_imp_NS[OF kbo_supt_one[OF this, of g Nil ss]] show ?thesis unfolding Cons by simp
  next
    case Nil
    from weight_w0[of "Fun g []"] have w: "weight (Fun g []) \<ge> weight (Fun f [])"
      using l[unfolded least] by auto
    from lex_ext_least_1
    have "snd (lex_ext kbo 0 [] [])" .
    hence lex: "snd (lex_ext_unbounded kbo [] [])" unfolding lex_ext_def Let_def by simp
    thus ?thesis using w l[unfolded least] unfolding Fun Nil by (auto simp: empty_le)
  qed
qed

lemma NS_least_least: assumes l: "least f"
  and NS: "NS (Fun f []) t"
  shows "\<exists> g. t = Fun g [] \<and> least g"
proof (cases t)
  case (Var x)
  show ?thesis using NS unfolding Var by simp
next
  case (Fun g ts)
  from NS[unfolded Fun, simplified]
  have w: "w (g,length ts) + listsum (map weight ts) \<le> weight (Fun f [])"
    by (auto split: if_splits)
  show ?thesis
  proof (cases ts)
    case Nil
    with w have "w (g,0) \<le> weight (Fun f [])" by simp
    also have "weight (Fun f []) \<le> w0" using l[unfolded least] by simp
    finally have g: "w (g,0) = w0" using  w0(1)[of g] by auto
    with w Nil l[unfolded least] have gf: "w (g,0) = w (f,0)" by simp
    with NS[unfolded Fun] have p: "prc (g,0) \<le> prc (f,0)" unfolding Nil 
      by (simp split: if_splits)
    have least: "least g" unfolding least
    proof (rule conjI[OF g], intro allI)
      fix h
      from l[unfolded least] have "w (h,0) = w0 \<longrightarrow> prc (f,0) \<le> prc (h,0)" by blast
      with p show "w (h,0) = w0 \<longrightarrow> prc (g,0) \<le> prc (h,0)" by auto
    qed
    show ?thesis 
      by (rule exI[of _ g], unfold Fun Nil, insert least, auto)
  next
    case (Cons s ss)
    from w l[unfolded least] have "w (g,length ts) + weight s + listsum (map weight ss) \<le> w0" unfolding Cons by simp
    with weight_w0[of s] 
    have wg: "w (g, length ts) = 0" and wss: "listsum (map weight ss) = 0" by arith+
    from wss weight_w0[of "hd ss"] w0(2) have ss: "ss = []" by (cases ss, auto)
    with Cons have ts: "ts = [s]" by simp
    from wg ts have wg: "w (g,1) = 0" by auto
    from adm[OF wg, rule_format, of f] have "prc (f,0) \<le> prc (g,1)" by auto
    with NS[unfolded Fun ts] l[unfolded least] weight_w0[of s] 
    have "snd (lex_ext_unbounded kbo [] [s])" 
      by (auto split: if_splits)        
    thus ?thesis unfolding lex_ext_unbounded_iff snd_conv by auto
  qed
qed

lemma not_S_least: assumes l: "least f"
  shows "\<not> S (Fun f []) t"
proof (cases t)
  case (Var x)
  show ?thesis unfolding Var by simp
next
  case (Fun g ts)
  show ?thesis unfolding Fun
  proof 
    assume S: "S (Fun f []) (Fun g ts)"
    from S[unfolded Fun, simplified]
    have w: "w (g,length ts) + listsum (map weight ts) \<le> weight (Fun f [])"
      by (auto split: if_splits)
    show False
    proof (cases ts)
      case Nil
      with w have "w (g,0) \<le> weight (Fun f [])" by simp
      also have "weight (Fun f []) \<le> w0" using l[unfolded least] by simp
      finally have g: "w (g,0) = w0" using  w0(1)[of g] by auto
      with w Nil l[unfolded least] have gf: "w (g,0) = w (f,0)" by simp
      with S have p: "prc (g,0) \<le> prc (f,0)" unfolding Nil 
        by (simp split: if_splits)
      with l[unfolded least, THEN conjunct2, rule_format, OF g] have p: "prc (g,0) = prc (f,0)" by force
      from p gf S have "fst (lex_ext_unbounded kbo [] ts)" unfolding Nil
        by auto
      thus False unfolding lex_ext_unbounded_iff by auto
    next
      case (Cons s ss)
      from w l[unfolded least] have "w (g,length ts) + weight s + listsum (map weight ss) \<le> w0" unfolding Cons by simp
      with weight_w0[of s] 
      have wg: "w (g, length ts) = 0" and wss: "listsum (map weight ss) = 0" by arith+
      from wss weight_w0[of "hd ss"] w0(2) have ss: "ss = []" by (cases ss, auto)
      with Cons have ts: "ts = [s]" by simp
      from wg ts have wg: "w (g,1) = 0" by auto
      from adm[OF wg, rule_format, of f] have "prc (f,0) \<le> prc (g,1)" by auto
      with S[unfolded Fun ts] l[unfolded least] weight_w0[of s] 
      have "fst (lex_ext_unbounded kbo [] [s])" 
        by (auto split: if_splits)    
      thus ?thesis unfolding lex_ext_unbounded_iff by auto
    qed
  qed
qed

lemma kbo_stable: fixes \<sigma> :: "('f,'v)subst" shows
  "(S s t \<longrightarrow> S (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)) \<and> (NS s t \<longrightarrow> NS (s \<cdot> \<sigma>) (t \<cdot> \<sigma>))" (is "?P s t")
proof (induct s t rule: kbo.induct)
  case (1 s t)
  show ?case
  proof (cases s)
    case (Var y)
    hence not: "\<not> S s t" using not_S_Var[of y t] by auto
    {
      assume "NS s t"
      from NS_Var_imp_eq_least[OF this[unfolded Var]]
      have "t = s \<or> (\<exists> f. t = Fun f [] \<and> least f)" unfolding Var  by simp
      then obtain f where "t = s \<or> t = Fun f [] \<and> least f" by auto
      hence "NS (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
      proof
        assume "t = s"
        thus ?thesis using NS_refl[of "t \<cdot> \<sigma>"] by auto
      next
        assume "t = Fun f [] \<and> least f"
        with NS_all_least[of f "s \<cdot> \<sigma>"] show ?thesis by auto
      qed
    }
    with not show ?thesis by blast
  next
    case (Fun f ss) note s = this
    let ?ss = "map (\<lambda> s. s \<cdot> \<sigma>) ss"
    show ?thesis
    proof (cases "NS s t")
      case False
      with S_imp_NS[of s t] show ?thesis by blast
    next
      case True note NS = this
      hence v: "vars_term_ms t \<le> vars_term_ms s" and w: "weight t \<le> weight s"
        by (auto split: if_splits)
      from weight_stable_le[OF w v] have w\<sigma>: "weight (t \<cdot> \<sigma>) \<le> weight (s \<cdot> \<sigma>)" by auto
      from vars_term_ms_subst_mono[OF v] have v\<sigma>: "vars_term_ms (t \<cdot> \<sigma>) \<le> vars_term_ms (s \<cdot> \<sigma>)" .
      show ?thesis
      proof (cases "weight (t \<cdot> \<sigma>) < weight (s \<cdot> \<sigma>)")
        case True
        with v\<sigma> show ?thesis by auto
      next
        case False
        with weight_stable_lt[OF _ v, of \<sigma>] w have w: "weight t = weight s" by arith
        show ?thesis
        proof (cases t)
          case (Var y)
          from set_of_mono[OF v, simplified]
          have "y \<in> vars_term s" unfolding Var by simp
          from supteq_Var[OF this] have "s \<rhd> Var y" unfolding Fun by auto 
          from kbo_supt[OF supt_stable[OF this]] have S: "S (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" unfolding Var .
          from S_imp_NS[OF S] S show ?thesis by auto
        next
          case (Fun g ts) note t = this
          let ?f = "(f,length ss)"
          let ?g = "(g,length ts)"
          let ?pf = "prc ?f"
          let ?pg = "prc ?g"          
          let ?ts = "map (\<lambda> s. s \<cdot> \<sigma>) ts"
          show ?thesis
          proof (cases "?pf > ?pg")
            case True 
            hence S: "S (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)" using w\<sigma> v\<sigma> unfolding s t by simp
            from S S_imp_NS[OF S] show ?thesis by simp
          next
            case False note prec = this
            show ?thesis
            proof (cases "?pf \<ge> ?pg")
              case False
              with v w prec have "\<not> NS s t" unfolding s t by (auto simp del: vars_term_ms.simps)
              with S_imp_NS[of s t] show ?thesis by blast
            next
              case True
              from v w have "vars_term_ms t \<le> vars_term_ms s \<and> weight t \<le> weight s" "\<not> weight t < weight s" by auto
              note IH = 1[OF this s t prec True]
              {
                assume "S s t"
                with prec v w True have lex: "fst (lex_ext_unbounded kbo ss ts)"
                  unfolding s t by simp
                have "fst (lex_ext_unbounded kbo ?ss ?ts)"
                  by (rule lex_ext_unbounded_map_S[OF _ _ lex], insert IH, auto)
                with v\<sigma> w\<sigma> prec True have "S (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
                  unfolding s t by auto
              }
              moreover 
              {
                assume "NS s t"
                with prec v w True have lex: "snd (lex_ext_unbounded kbo ss ts)"
                  unfolding s t by simp
                have "snd (lex_ext_unbounded kbo ?ss ?ts)"
                  by (rule lex_ext_unbounded_map_NS[OF _ _ lex], insert IH, auto)
                with v\<sigma> w\<sigma> prec True have "NS (s \<cdot> \<sigma>) (t \<cdot> \<sigma>)"
                  unfolding s t by auto
              }
              ultimately show ?thesis by auto
            qed
          qed
        qed
      qed
    qed
  qed
qed

lemma S_subst: "S s t \<Longrightarrow> S (s \<cdot> (\<sigma> :: ('f,'v)subst)) (t \<cdot> \<sigma>)" using kbo_stable[of s t \<sigma>] by auto
lemma NS_subst: "NS s t \<Longrightarrow> NS (s \<cdot> (\<sigma> :: ('f,'v)subst)) (t \<cdot> \<sigma>)" using kbo_stable[of s t \<sigma>] by auto

lemma kbo_trans: "(S s t \<longrightarrow> NS t u \<longrightarrow> S s u) \<and> (NS s t \<longrightarrow> S t u \<longrightarrow> S s u) \<and> (NS s t \<longrightarrow> NS t u \<longrightarrow> NS s u)" (is "?P s t u")
proof (induct s arbitrary: t u)
  case (Var x t u)
  from not_S_Var[of x t] have nS: "\<not> S (Var x) t" .
  show ?case
  proof (cases "NS (Var x) t")
    case False
    with nS show ?thesis by blast
  next
    case True
    from NS_Var_imp_eq_least[OF this] obtain f where 
      "t = Var x \<or> t = Fun f [] \<and> least f" by blast
    thus ?thesis
    proof 
      assume "t = Var x"
      thus ?thesis using nS by blast
    next
      assume "t = Fun f [] \<and> least f"
      hence t: "t = Fun f []" and least: "least f" by auto
      from not_S_least[OF least] have nS': "\<not> S t u" unfolding t .
      show ?thesis
      proof (cases "NS t u")
        case True
        with NS_least_least[OF least, of u] t obtain h where 
          u: "u = Fun h []" and least: "least h" by auto
        from NS_all_least[OF least] have NS: "NS (Var x) u" unfolding u .
        with nS nS' show ?thesis by blast
      next
        case False
        with S_imp_NS[of t u] show ?thesis by blast
      qed
    qed
  qed   
next
  case (Fun f ss t u) note IH = this
  let ?s = "Fun f ss"
  show ?case
  proof (cases "NS ?s t")
    case False
    with S_imp_NS[of ?s t] show ?thesis by blast
  next
    case True note st = this
    hence vst: "vars_term_ms t \<le> vars_term_ms ?s" and wst: "weight t \<le> weight ?s"
      by (auto split: if_splits)
    show ?thesis
    proof (cases "NS t u")
      case False
      with S_imp_NS[of t u] show ?thesis by blast
    next
      case True note tu = this
      hence vtu: "vars_term_ms u \<le> vars_term_ms t" and wtu: "weight u \<le> weight t"
        by (auto split: if_splits)
      from vst vtu have v: "vars_term_ms u \<le> vars_term_ms ?s" by simp
      from wst wtu have w: "weight u \<le> weight ?s" by simp
      show ?thesis
      proof (cases "weight u < weight ?s")
        case True
        with v show ?thesis by auto
      next
        case False
        with wst wtu have wst: "weight t = weight ?s" and wtu: "weight u = weight t" and w: "weight u = weight ?s" by arith+                
        show ?thesis
        proof (cases u)
          case (Var z)
          with v w show ?thesis by auto
        next
          case (Fun h us) note u = this
          show ?thesis
          proof (cases t)
            case (Fun g ts) note t = this
            let ?f = "(f,length ss)"
            let ?g = "(g,length ts)"
            let ?h = "(h,length us)"
            let ?pf = "prc ?f"
            let ?pg = "prc ?g"
            let ?ph = "prc ?h"            
            from st t wst have fg: "?pg \<le> ?pf" by (simp split: if_splits)
            from tu t u wtu have gh: "?ph \<le> ?pg" by (simp split: if_splits)
            from fg gh have fh: "?ph \<le> ?pf" by simp          
            show ?thesis
            proof (cases "?ph < ?pf")
              case True
              with w v u show ?thesis by auto
            next
              case False
              let ?lex = "lex_ext_unbounded kbo"
              from False fg gh have fg: "?pg = ?pf" and gh: "?ph = ?pg" and fh: "?ph = ?pf" by arith+
              from st t wst fg have g: "?pg = ?pf" and st: "snd (?lex ss ts)"
                by (auto split: if_splits)
              from tu t u wtu gh have h: "?ph = ?pg" and tu: "snd (?lex ts us)" 
                by (auto split: if_splits)
              from g h have hf: "?ph = ?pf" by simp
              {
                fix s t u
                assume "s \<in> set ss"
                from IH[OF this, of t u]
                have "(NS s t \<and> S t u \<longrightarrow> S s u) \<and> 
                  (S s t \<and> NS t u \<longrightarrow> S s u) \<and>
                  (NS s t \<and> NS t u \<longrightarrow> NS s u) \<and> 
                  (S s t \<and> S t u \<longrightarrow> S s u)"
                  using S_imp_NS[of s t] by blast
              } note IH = this
              let ?b = "length ss + length ts + length us"
              note lex = lex_ext_compat[of ss ts us kbo ?b, OF IH]
              let ?lexb = "lex_ext kbo ?b"
              note conv = lex_ext_def Let_def
              from st have st: "snd (?lexb ss ts)" unfolding conv by simp
              from tu have tu: "snd (?lexb ts us)" unfolding conv by simp
              from lex st tu have su: "snd (?lexb ss us)" by blast
              hence su: "snd (?lex ss us)" unfolding conv by simp
              from w v u su hf have NS: "NS ?s u" by simp
              {
                assume st: "S ?s t"
                with t wst fg have st: "fst (?lex ss ts)"
                  by (auto split: if_splits)
                hence st: "fst (?lexb ss ts)" unfolding conv by simp
                from lex st tu have su: "fst (?lexb ss us)" by blast
                hence su: "fst (?lex ss us)" unfolding conv by simp
                from w v u su hf have S: "S ?s u" by simp
              } note S_left = this
              {
                assume tu: "S t u"
                with t u wtu gh have tu: "fst (?lex ts us)"
                  by (auto split: if_splits)
                hence tu: "fst (?lexb ts us)" unfolding conv by simp
                from lex st tu have su: "fst (?lexb ss us)" by blast
                hence su: "fst (?lex ss us)" unfolding conv by simp
                from w v u su hf have S: "S ?s u" by simp
              } note S_right = this
              from NS S_left S_right show ?thesis by blast
            qed
          next
            case (Var x) note t = this
            from tu weight_w0[of u] have least: "least h" and u: "u = Fun h []" unfolding t u
              by (auto split: if_splits)
            from NS_all_least[OF least] have NS: "NS ?s u" unfolding u .
            from not_S_Var have nS': "\<not> S t u" unfolding t .
            show ?thesis
            proof (cases "S ?s t")
              case False
              with nS' NS show ?thesis by blast
            next
              case True
              hence "vars_term_ms t \<le> vars_term_ms ?s" 
                by (auto split: if_splits)
              from set_of_mono[OF this, unfolded set_of_vars_term_ms t]
              have "x \<in> vars_term ?s" by simp
              then obtain s sss where ss: "ss = s # sss" by (cases ss, auto)
              from kbo_supt_one[OF NS_all_least[OF least, of s], of f Nil sss]
              have "S ?s u" unfolding ss u by simp
              with NS show ?thesis by blast
            qed
          qed
        qed 
      qed
    qed
  qed
qed

lemma S_trans: "S s t \<Longrightarrow> S t u \<Longrightarrow> S s u" using S_imp_NS[of s t] kbo_trans[of s t u] by blast
lemma NS_trans: "NS s t \<Longrightarrow> NS t u \<Longrightarrow> NS s u" using kbo_trans[of s t u] by blast
lemma NS_S_compat: "NS s t \<Longrightarrow> S t u \<Longrightarrow> S s u" using kbo_trans[of s t u] by blast
lemma S_NS_compat: "S s t \<Longrightarrow> NS t u \<Longrightarrow> S s u" using kbo_trans[of s t u] by blast

lemma kbo_strongly_normalizing: fixes s :: "('f,'v)term" 
  shows "SN_on {(s,t). S s t} {s}"
proof -
  let ?S = "{(s,t). S s t}"
  let ?SN = "\<lambda> t :: ('f,'v)term. SN_on ?S {t}"
  let ?m1 = "\<lambda> (f,ss). weight (Fun f ss)"
  let ?m2 = "\<lambda> (f,ss). prc (f,length ss)"
  let ?ms = "[?m1,?m2]"
  let ?m = "measures ?ms"
  note conv = SN_on_all_reducts_SN_on_conv[of ?S]
  show "?SN s"
  proof (induct s)
    case (Var x)
    show ?case unfolding conv[of "Var x"] using not_S_Var[of x] by auto
  next
    case (Fun f ss)
    hence subset: "set ss \<subseteq> {s. ?SN s}" by blast
    let ?P = "\<lambda> (f,ss). set ss \<subseteq> {s. ?SN s} \<longrightarrow> ?SN (Fun f ss)"
    {
      fix fss
      have "?P fss"
      proof (induct fss rule: wf_induct[OF wf_measures[of ?ms]])
        case (1 fss)
        obtain f ss where fss: "fss = (f,ss)" by force
        {
          fix g ts
          note IH = 1[rule_format, of "(g,ts)", unfolded fss split, THEN mp]
        } note IH = this
        show ?case unfolding fss split
        proof
          assume SN_s: "set ss \<subseteq> {s. ?SN s}"
          let ?f = "(f,length ss)"
          let ?s = "Fun f ss"
          let ?w = "weight ?s"
          let ?p = "prc ?f"
          let ?SNt = "\<lambda> g ts. ?SN (Fun g ts)"
          let ?sym = "\<lambda> g ts. (g,length ts)"
          let ?lex = "lex_ext kbo ?w"
          let ?lexu = "lex_ext_unbounded kbo"
          let ?lex_SN = "{(ys,xs). (\<forall> y \<in> set ys. ?SN y) \<and> fst (?lex ys xs)}"
          from lex_ext_SN[of kbo ?w, OF S_imp_NS NS_S_compat]
          have SN: "SN ?lex_SN" .
          {
            fix g and ts :: "('f,'v)term list"
            assume "prc (?sym g ts) \<le> ?p \<and> weight (Fun g ts) \<le> ?w \<and> set ts \<subseteq> {s. ?SN s}"
            hence "?SNt g ts"
            proof (induct ts arbitrary: g rule: SN_induct[OF SN])
              case (1 ts g)
              note inner_IH = 1(1)
              let ?g = "(g,length ts)"
              let ?t = "Fun g ts"
              from 1(2) have prc: "prc ?g \<le> prc ?f" and w: "weight ?t \<le> ?w"  and SN: "set ts \<subseteq> {s. ?SN s}" by auto
              show "?SNt g ts" unfolding conv[of ?t]
              proof (intro allI impI)
                fix u
                assume "(?t,u) \<in> ?S"
                hence tu: "S ?t u" by auto
                thus "?SN u"
                proof (induct u)
                  case (Var x)
                  thus ?case using not_S_Var[of x] unfolding conv[of "Var x"] by auto
                next
                  case (Fun h us) 
                  let ?h = "(h,length us)"
                  let ?u = "Fun h us"
                  note tu = Fun(2)
                  {
                    fix u 
                    assume u: "u \<in> set us"
                    hence "?u \<rhd> u" by auto
                    from S_trans[OF tu kbo_supt[OF this]] have "S ?t u" by auto
                    from Fun(1)[OF u this] have "?SN u" .
                  } hence SNu: "set us \<subseteq> {s . ?SN s}" by blast
                  note IH = IH[OF _ this]
                  from tu have wut: "weight ?u \<le> weight ?t" by (simp split: if_splits)
                  show ?case
                  proof (cases "weight ?u = weight ?t")
                    case False
                    with wut w have w: "weight ?u < weight ?s" by auto
                    show ?thesis
                      by (rule IH, insert w, auto)
                  next
                    case True
                    with w have w: "weight ?u \<le> weight ?s" by auto
                    note tu = tu[unfolded kbo.simps True, unfolded Fun term.simps, simplified]
                    show ?thesis 
                    proof (cases "prc ?h < prc ?g")
                      case True
                      with prc have prc: "prc ?h < prc ?f" by simp
                      show ?thesis
                        by (rule IH, insert prc w, auto)
                    next
                      case False
                      with tu have hg: "prc ?h = prc ?g" 
                        and lex: "fst (?lexu ts us)" by (auto split: if_splits)
                      with prc have prc_hf: "prc ?h \<le> prc ?f" by auto
                      from prc_hf w SNu have "prc ?h \<le> prc ?f \<and> weight ?u \<le> ?w \<and> set us \<subseteq> {s. ?SN s}" 
                        by auto
                      note inner_IH = inner_IH[OF _ this]
                      show ?thesis 
                      proof (rule inner_IH, rule, unfold split, intro conjI ballI)
                        have "fst (?lexu ts us)" by (rule lex)
                        moreover have "length us \<le> ?w" 
                        proof -
                          have "length us \<le> listsum (map weight us)" 
                          proof (induct us)
                            case (Cons u us)
                            from Cons have "length (u # us) \<le> Suc (listsum (map weight us))" by auto
                            also have "... \<le> listsum (map weight (u # us))" using weight_w0[of u] w0(2)
                              by auto
                            finally show ?case .
                          qed simp
                          also have "... \<le> ?w" using w by simp
                          finally show ?thesis .
                        qed                          
                        ultimately show "fst (?lex ts us)" unfolding lex_ext_def Let_def by auto
                      next
                        fix t
                        assume t: "t \<in> set ts" 
                        with SN show "?SN t" by blast
                      qed
                    qed
                  qed
                qed
              qed
            qed
          }
          from this[of f ss] SN_s  show "?SN ?s" by auto
        qed
      qed
    }
    from this[of "(f,ss)", unfolded split]
    show ?case using Fun by blast
  qed
qed

declare kbo.simps[simp del]

abbreviation kbo_S where "kbo_S \<equiv> {(s,t). S s t}"
abbreviation kbo_NS where "kbo_NS \<equiv> {(s,t). NS s t}"

lemma kbo_ce_compat: "ce_trs x \<subseteq> kbo_S \<inter> kbo_NS"
proof (cases x)
  case (fields a b c)
  {
    fix s t
    assume "(s,t) \<in> ce_trs x"
    hence "s \<rhd> t" unfolding fields ce_trs.simps by auto
    from kbo_supt[OF this] have "(s,t) \<in> kbo_S" by simp
    with S_imp_NS[of s t] have "(s,t) \<in> kbo_S \<inter> kbo_NS" by auto
  }
  thus ?thesis by auto
qed

lemma S_ce_compat: "ce_trs x \<subseteq> kbo_S" using kbo_ce_compat by blast
lemma NS_ce_compat: "ce_trs x \<subseteq> kbo_NS" using kbo_ce_compat by blast

lemma kbo_redpair: "mono_ce_af_redtriple_order kbo_S kbo_NS kbo_NS full_af"
proof (unfold_locales)
  show "SN kbo_S" using kbo_strongly_normalizing unfolding SN_defs by blast
  show "subst_closed kbo_S" using S_subst by auto
  show "subst_closed kbo_NS" using NS_subst by auto
  show "ctxt_closed kbo_NS" 
    by (rule one_imp_ctxt_closed, insert NS_mono_one, auto)
  show "kbo_S O kbo_NS \<subseteq> kbo_S" using S_NS_compat by auto
  show "kbo_NS O kbo_S \<subseteq> kbo_S" using NS_S_compat by auto
  show "trans kbo_S" using S_trans unfolding trans_def by auto
  show "trans kbo_NS" using NS_trans unfolding trans_def by auto
  show "refl kbo_NS" using NS_refl unfolding refl_on_def by auto
  show "af_compatible full_af kbo_NS" by (rule full_af)
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c d . ce_trs (c,d,m) \<subseteq> kbo_NS"
    using NS_ce_compat by blast
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c d . ce_trs (c,d,m) \<subseteq> kbo_S"
    using S_ce_compat by blast
qed

lemma kbo_S_mono: "ctxt_closed kbo_S"
  by (rule one_imp_ctxt_closed, insert S_mono_one, auto)

end

end